/*
 * Copyright 2015 Haruki Hasegawa
 * Copyright 2016 Christian Basler
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package ch.dissem.apps.abit.adapter

import android.content.Context
import android.graphics.*
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.FrameLayout
import android.widget.ImageView
import android.widget.TextView
import androidx.core.content.ContextCompat
import androidx.recyclerview.widget.ItemTouchHelper
import androidx.recyclerview.widget.RecyclerView
import ch.dissem.apps.abit.Identicon
import ch.dissem.apps.abit.MultiIdenticon
import ch.dissem.apps.abit.R
import ch.dissem.apps.abit.util.Strings.prepareMessageExtract
import ch.dissem.apps.abit.util.getDrawable
import ch.dissem.apps.abit.util.getString
import ch.dissem.bitmessage.entity.Conversation
import ch.dissem.bitmessage.entity.Plaintext
import ch.dissem.bitmessage.entity.valueobject.Label
import java.util.*

/**
 * Adapted from the basic swipeable example by Haruki Hasegawa. See
 *
 * @author Christian Basler
 * @see [https://github.com/h6ah4i/android-advancedrecyclerview](https://github.com/h6ah4i/android-advancedrecyclerview)
 */
abstract class SwipeableAdapter<T, H>(val ctx: Context) :
    RecyclerView.Adapter<H>() where H : SwipeableAdapter.AbstractViewHolder {

    protected val data = LinkedList<T>()
    var eventListener: EventListener? = null

    protected var label: Label? = null
    var selectedPosition = -1
        set(value) {
            val oldPosition = field
            field = value
            notifyItemChanged(oldPosition)
            notifyItemChanged(value)
        }
    var activateOnItemClick: Boolean = false

    protected val labelUnknown: String = ctx.getString(R.string.unknown)

    open class AbstractViewHolder(v: View, adapter: SwipeableAdapter<*, *>) : RecyclerView.ViewHolder(v) {

        val container = v.findViewById<FrameLayout>(R.id.container)!!

        init {
            itemView.setOnClickListener { adapter.eventListener?.onItemSelected(adapterPosition) }
            container.setOnClickListener { adapter.eventListener?.onItemSelected(adapterPosition) }
        }
    }

    init {
        // SwipeableItemAdapter requires stable ID, and also
        // have to implement the getItemId() method appropriately.
        setHasStableIds(true)
    }

    override fun onBindViewHolder(holder: H, position: Int) {
        val item = data[position]

        holder.apply {
            if (activateOnItemClick) {
                container.setBackgroundResource(
                    if (position == selectedPosition)
                        R.drawable.bg_item_selected_state
                    else
                        R.drawable.bg_item_normal_state
                )
            }

            setData(holder, item)
        }
    }

    abstract fun setData(holder: H, item: T)

    fun add(item: T) {
        val index = data.size
        data.add(item)
        notifyItemInserted(index)
    }

    fun addFirst(item: T) {
        data.addFirst(item)
        notifyItemInserted(0)
    }

    fun addAll(items: Collection<T>) {
        val index = data.size
        data.addAll(items)
        notifyItemRangeInserted(index, items.size)
    }

    fun remove(item: T) {
        val itemId = getItemId(item)
        val index = data.indexOfFirst { getItemId(it) == itemId }
        if (index >= 0) {
            removeAt(index)
        }
    }

    fun removeAt(index: Int) {
        data.removeAt(index)
        notifyItemRemoved(index)
    }

    override fun getItemId(position: Int): Long {
        return getItemId(data[position])
    }

    abstract fun getItemId(item: T): Long

    abstract fun update(item: T)

    fun clear(newLabel: Label?) {
        label = newLabel
        data.clear()
        notifyDataSetChanged()
    }

    fun getItem(position: Int) = data[position]

    override fun getItemCount() = data.size

}

class SwipeableConversationAdapter(ctx: Context) : SwipeableAdapter<Conversation, SwipeableConversationAdapter.ViewHolder>(ctx) {
    override fun getItemId(item: Conversation) = item.id.leastSignificantBits

    override fun update(item: Conversation) {
        val index = data.indexOfFirst { it.id == item.id }
        if (index >= 0) {
            data[index] = item
            notifyItemChanged(index)
        }
    }

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): ViewHolder {
        val inflater = LayoutInflater.from(parent.context)
        val v = inflater.inflate(R.layout.conversation_row, parent, false)
        return ViewHolder(v, this)
    }

    override fun setData(holder: ViewHolder, item: Conversation) {
        holder.apply {
            avatar.setImageDrawable(MultiIdenticon(item.participants))

            sender.text = item.participants.sortedBy {
                (it.alias?.let { 0 } ?: 1) + if (it.isChan) 2 else 0
            }.map { it.alias ?: labelUnknown }.distinct().joinToString()
            subject.text = prepareMessageExtract(item.subject)
            extract.text = prepareMessageExtract(item.extract)
            item.messages.count { it.labels.contains(label) }.let { size ->
                if (size <= 1) {
                    count.text = ""
                } else {
                    count.text = size.toString()
                }
            }
            if (item.hasUnread()) {
                sender.typeface = Typeface.DEFAULT_BOLD
                subject.typeface = Typeface.DEFAULT_BOLD
            } else {
                sender.typeface = Typeface.DEFAULT
                subject.typeface = Typeface.DEFAULT
            }
        }
    }

    class ViewHolder(v: View, adapter: SwipeableConversationAdapter) : AbstractViewHolder(v, adapter) {
        val avatar = v.findViewById<ImageView>(R.id.avatar)!!
        val status = v.findViewById<ImageView>(R.id.status)!!
        val sender = v.findViewById<TextView>(R.id.sender)!!
        val subject = v.findViewById<TextView>(R.id.subject)!!
        val extract = v.findViewById<TextView>(R.id.text)!!
        val count = v.findViewById<TextView>(R.id.count)!!
    }
}

class SwipeableMessageAdapter(ctx: Context) : SwipeableAdapter<Plaintext, SwipeableMessageAdapter.ViewHolder>(ctx) {
    override fun getItemId(item: Plaintext) = item.id as Long

    override fun update(item: Plaintext) {
        val index = data.indexOfFirst { it.id == item.id }
        if (index >= 0) {
            data[index] = item
            notifyItemChanged(index)
        }
    }

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): ViewHolder {
        val inflater = LayoutInflater.from(parent.context)
        val v = inflater.inflate(R.layout.message_row, parent, false)
        return ViewHolder(v, this)
    }

    override fun setData(holder: ViewHolder, item: Plaintext) {
        holder.apply {
            avatar.setImageDrawable(Identicon(item.from))
            status.setImageResource(item.status.getDrawable())
            status.contentDescription = holder.status.context.getString(item.status.getString())

            sender.text = item.from.toString()
            subject.text = prepareMessageExtract(item.subject)
            extract.text = prepareMessageExtract(item.text)
            if (item.isUnread()) {
                sender.typeface = Typeface.DEFAULT_BOLD
                subject.typeface = Typeface.DEFAULT_BOLD
            } else {
                sender.typeface = Typeface.DEFAULT
                subject.typeface = Typeface.DEFAULT
            }
        }
    }

    class ViewHolder(v: View, adapter: SwipeableMessageAdapter) : AbstractViewHolder(v, adapter) {
        val avatar = v.findViewById<ImageView>(R.id.avatar)!!
        val status = v.findViewById<ImageView>(R.id.status)!!
        val sender = v.findViewById<TextView>(R.id.sender)!!
        val subject = v.findViewById<TextView>(R.id.subject)!!
        val extract = v.findViewById<TextView>(R.id.text)!!
    }
}

class SwipeToDeleteCallback(ctx: Context, swipeDirs: Int, private val eventListener: EventListener) : ItemTouchHelper.SimpleCallback(0, swipeDirs) {

    private val backgroundLeft = ContextCompat.getDrawable(ctx, R.drawable.bg_swipe_item_left)!!
    private val backgroundRight = ContextCompat.getDrawable(ctx, R.drawable.bg_swipe_item_right)!!
    private val clearPaint = Paint().apply { xfermode = PorterDuffXfermode(PorterDuff.Mode.CLEAR) }


    override fun getMovementFlags(recyclerView: RecyclerView, viewHolder: RecyclerView.ViewHolder): Int {
        /**
         * To disable "swipe" for specific item return 0 here.
         * For example:
         * if (viewHolder?.itemViewType == YourAdapter.SOME_TYPE) return 0
         * if (viewHolder?.adapterPosition == 0) return 0
         */
        if (viewHolder.adapterPosition == 10) return 0
        return super.getMovementFlags(recyclerView, viewHolder)
    }

    override fun onMove(recyclerView: RecyclerView, viewHolder: RecyclerView.ViewHolder, target: RecyclerView.ViewHolder) = false

    override fun onChildDraw(
        c: Canvas, recyclerView: RecyclerView, viewHolder: RecyclerView.ViewHolder,
        dX: Float, dY: Float, actionState: Int, isCurrentlyActive: Boolean
    ) {

        val itemView = viewHolder.itemView
        val isCanceled = dX == 0f && !isCurrentlyActive

        if (isCanceled) {
            clearCanvas(c, itemView.right + dX, itemView.top.toFloat(), itemView.right.toFloat(), itemView.bottom.toFloat())
            super.onChildDraw(c, recyclerView, viewHolder, dX, dY, actionState, isCurrentlyActive)
            return
        }

        if (dX < 0) {
            backgroundLeft.setBounds(itemView.right + dX.toInt(), itemView.top, itemView.right, itemView.bottom)
            backgroundLeft.draw(c)
        } else {
            backgroundRight.setBounds(itemView.left, itemView.top, itemView.left + dX.toInt(), itemView.bottom)
            backgroundRight.draw(c)
        }

        super.onChildDraw(c, recyclerView, viewHolder, dX, dY, actionState, isCurrentlyActive)
    }

    private fun clearCanvas(c: Canvas?, left: Float, top: Float, right: Float, bottom: Float) {
        c?.drawRect(left, top, right, bottom, clearPaint)
    }

    override fun onSwiped(viewHolder: RecyclerView.ViewHolder, direction: Int) {
        when (direction) {
            ItemTouchHelper.LEFT -> eventListener.onItemDeleted(viewHolder.adapterPosition)
            ItemTouchHelper.RIGHT -> eventListener.onItemArchived(viewHolder.adapterPosition)
        }
    }
}

interface EventListener {
    fun onItemDeleted(position: Int)

    fun onItemArchived(position: Int)

    fun onItemSelected(position: Int)
}
